#pragma once

#include <iostream>
#include <iomanip>
#include <vector>
#include <array>

// узел
struct Node {
    int id;
    std::array<double, 3> coord; 
    bool ver;

    Node() = default;
    Node(int, std::array<double, 3>, bool);

    bool operator==(const Node&) const;
    bool operator<(const Node&) const;

    friend std::ostream& operator<<(std::ostream&, const Node&);
    friend std::ostream& operator<<(std::ostream&, const std::vector<Node>&);
};

// КЭ 
struct FiniteElement {
    int elem_id;
    int material_id;
    std::vector<int> nodes_id;

    friend std::ostream& operator<<(std::ostream&, const FiniteElement&);
    friend std::ostream& operator<<(std::ostream&, const std::vector<FiniteElement>&);
};

// ПКЭ
struct BoundaryFiniteElement : public FiniteElement {
    int elem_id;
    int rib_id;
    std::vector<int> nodes_id;

    friend std::ostream& operator<<(std::ostream&, const BoundaryFiniteElement&);
    friend std::ostream& operator<<(std::ostream&, const std::vector<BoundaryFiniteElement>&);
};

// ребро
struct Rib {
    int first_node_id;
    int last_node_id;
    int center_id;

    Rib() = default;
    Rib(int, int, int);

    bool operator==(const Rib&) const;
    bool operator<(const Rib&) const;
    void UpdateCenter(int);
};

bool Node::operator==(const Node& right_node) const {
    return
        (id == right_node.id && 
         coord == right_node.coord &&
         ver == right_node.ver);
};

std::ostream& operator<<(std::ostream& out_stream, const Node& node_) {
    out_stream <<
        std::setw(4) << node_.id << "  " <<
        std::right << std::setw(11) <<
        node_.ver << std::endl;
    for (const auto& elem : node_.coord) {
        out_stream << elem << " ";
    }

    return out_stream;
}

std::ostream& operator<<(std::ostream& out_stream, const FiniteElement& element_) {
    out_stream <<
        std::setw(4) <<
        element_.elem_id << "  " <<
        std::setw(11) <<
        element_.material_id << "  ";

    for (const auto& nodes_id_it : element_.nodes_id) {
        out_stream << std::right << std::setw(4) << nodes_id_it;
    }

    out_stream << std::endl;

    return out_stream;
}

std::ostream& operator<<(std::ostream& out_stream, const BoundaryFiniteElement& surface_) {
    out_stream <<
        std::setw(9) << surface_.elem_id << "  " <<
        std::setw(9) << surface_.rib_id << "  ";

    for (const auto& nodes_id_it : surface_.nodes_id) {
        out_stream << std::right << std::setw(4) << nodes_id_it;
    }

    out_stream << std::endl;

    return out_stream;
}

std::ostream& operator<<(std::ostream& out_stream, const std::vector<Node>& nodes_) {
    out_stream << nodes_.size() << (nodes_.size() > 1 ? " Nodes:\n" : " Node:\n") << " id"
        << std::setw(10) << "x" << std::setw(11) << "y" << std::setw(11) << "z" << std::setw(14) << "Vertex\n";
    for (const auto& it : nodes_) {
        out_stream << it;
    }
    return out_stream;
}

Node::Node(int id_, std::array<double, 3> coord_, bool ver_) :
    id(id_), coord(coord_), ver(ver_) {}

bool Node::operator<(const Node& right_node) const {
    return id < right_node.id;
}

std::ostream& operator<<(std::ostream& out_stream, const std::vector<FiniteElement>& elements_) {
    int elements_size = elements_.size();
    out_stream <<
        elements_size <<
        (elements_size > 1 ? " Elements:\n" : " Element:\n") << " id" <<
        std::setw(15) << "Material_id" << std::setw(15) << "Nodes_id" << std::endl;

    for (const auto& it : elements_) {
        out_stream << it;
    }

    return out_stream;
}

std::ostream& operator<<(std::ostream& out_stream, const std::vector<BoundaryFiniteElement>& elem_) {
    int elem_size = elem_.size();

    out_stream <<
        elem_size <<
        (elem_size > 1 ? " BoundaryFiniteElement:\n" : " BoundaryFiniteElement:\n") << "BoundaryFiniteElement_ID" <<
        std::setw(11) << "Border_id" << std::setw(14) << "Nodes_id\n";

    for (const auto& it : elem_) {
        out_stream << it;
    }

    return out_stream;
}

bool Rib::operator==(const Rib& right_edge) const {
    if (right_edge.first_node_id == first_node_id && right_edge.last_node_id == last_node_id) {
        return true;
    }
    else {
        return (right_edge.first_node_id == last_node_id && right_edge.last_node_id == first_node_id);
    }
}

Rib::Rib(int id1, int id2, int id3) :
    first_node_id(id1), last_node_id(id2), center_id(id3) {}

void Rib::UpdateCenter(int id_) {
    center_id = id_;
}

bool Rib::operator<(const Rib& right_edge) const {
    return (center_id < right_edge.center_id);
}
